#include <Date.au3>
#include <GUIConstants.au3>
#include <WinAPI.au3>
#include <WinAPIEx.au3>
#include <WindowsConstants.au3>

#include "TetrisGraphics.au3"



Local Const $TETRIS_TIMEOUT = 500

Local Const $DATABASE_FILENAME = @WorkingDir & "\TETRIS_LOG.MDB"
Local Const $DATABASE_PASSWORD = "thornbird313@126.com"
Local Const $ACCESS_CONNECTION = "Provider=Microsoft.Jet.OLEDB.4.0;" & _
		"Data Source=" & $DATABASE_FILENAME & ";" & _
		"Jet OLEDB:Database Password=" & $DATABASE_PASSWORD

Local $hCurrentWnd = 0
Local $hKeyboardHook = 0
Local $aLayout[18][10]
Local $iCurrentShapeIndex = 0
Local $iCurrentShapeDirection = 0
Local $aCurrentShapePosition[4][4]
Local $iCurrentShapeX = 0
Local $iCurrentShapeY = 0
Local $iShapeIndex = 0
Local $iSeconds = 0
Local $iRowCount = 0
Local $iScore = 0
Local $iTotalSeconds = 0
Local $fRunning = False
Local $fGameOver = False
Local $iStartTickCount = 0
Local $hTimerProcHandle = DllCallbackRegister("TimerProc", "uint", "hwnd;uint;uint;dword")
Local $fTimerRunning = False
Local $iTimerID = 0
Local $fKeyDownFlag = False


Func Initialize($hWnd)
	Local $hKeyboardHandler, $hModule
	$hCurrentWnd = $hWnd
	FileInstall("Tetris.bmp", $TETRIS_LOGO_FILENAME, 1)
	FileInstall("TetrisHelp.bmp", $TETRIS_HELP_FILENAME, 1)
	InitializeGameData()
	GUIRegisterMsg($WM_PAINT, "DrawTetris")
	$hKeyboardHandler = DllCallbackRegister("KeyboardHandler", "long", "int;wparam;lparam")
	$hModule = _WinAPI_GetModuleHandle(0) 
	$hKeyboardHook = _WinAPI_SetWindowsHookEx($WH_KEYBOARD_LL, DllCallbackGetPtr($hKeyboardHandler), $hModule)
EndFunc


Func Destroy()
	StopTimer()
	DllCallbackFree($hTimerProcHandle)
	DeleteObjects()
	FileDelete($TETRIS_LOGO_FILENAME)
	FileDelete($TETRIS_HELP_FILENAME)
EndFunc


Func InitializeGameData()
	Local $i, $j
	For $i = 0 To 17
		For $j = 0 To 9
			$aLayout[$i][$j] = 0
		Next
	Next
	$iCurrentShapeIndex = 0
	$iCurrentShapeDirection = 0
	For $i = 0 To 3
		For $j = 0 To 3
			$aCurrentShapePosition[$i][$j] = 0
		Next
	Next
	$iCurrentShapeX = 0
	$iCurrentShapeY = 0
	$iShapeIndex = 0
	$iSeconds = 0
	$iRowCount = 0
	$iScore = 0
	$iTotalSeconds = 0
	$fRunning = False
	$fGameOver = False
EndFunc


Func StartGame()
	If $fRunning = False Then
		StartTimer()
		$fRunning = True
	EndIf
EndFunc


Func StopGame()
	If $fRunning = True Then
		StopTimer()
		$fRunning = False
	EndIf
EndFunc


Func GameOver()
	Local $sName
	StopGame()
	$fGameOver = True
	$iTotalSeconds = Round((_Date_Time_GetTickCount() - $iStartTickCount) / 1000, 0)
	$sName = StringStripWS(InputBox("ɼ", "", ""), 3)
	If StringLen($sName) = 0 Then Return
	SaveResult($sName)
EndFunc


Func StartTimer()
	StopTimer()
	_WinAPI_SetTimer($hCurrentWnd, $iTimerID, $TETRIS_TIMEOUT, DllCallbackGetPtr($hTimerProcHandle))
	$iStartTickCount = _Date_Time_GetTickCount()
	GenerateNewShape()
	$fTimerRunning = True
	DrawTetris()
EndFunc


Func StopTimer()
	If $fTimerRunning = True Then
		_WinAPI_KillTimer($hCurrentWnd, $iTimerID)
		$fTimerRunning = False
		$iTimerID = 0
		DrawTetris()
	EndIf
EndFunc


Func TimerProc($hWnd, $iMsg, $wParam, $lParam)
	If $fKeyDownFlag = True Then Return
	$iSeconds = Round((_Date_Time_GetTickCount() - $iStartTickCount) / 1000, 0)
	MoveDown()
EndFunc


Func DrawTetris($hWnd = 0, $iMsg = 0, $wParam = 0, $lParam = 0)
	DrawTetrisGraphics($hCurrentWnd, $aLayout, $aCurrentShapePosition, $iCurrentShapeIndex, _
			$iCurrentShapeX, $iCurrentShapeY, $iShapeIndex, $iSeconds, $iRowCount, $iScore, _
			$fGameOver)	
    Return $GUI_RUNDEFMSG
EndFunc


Func KeyboardHandler($nCode, $wParam, $lParam)
	Local $tKeyHooks, $iKeyCode
	$tKeyHooks = DllStructCreate($tagKBDLLHOOKSTRUCT, $lParam)
	If $nCode < 0 Then
		Return _WinAPI_CallNextHookEx($hKeyboardHook, $nCode, $wParam, $lParam)
	EndIf
	If $wParam = $WM_KEYDOWN Then
		$iKeyCode = DllStructGetData($tKeyHooks, "vkCode")
		If $fRunning = True And ($iKeyCode = 38 Or $iKeyCode = 40 Or $iKeyCode = 37 _
				Or $iKeyCode = 39 Or $iKeyCode = 32) Then
			OnKeyDown($iKeyCode)
		EndIf
	EndIf
	Return _WinAPI_CallNextHookEx($hKeyboardHook, $nCode, $wParam, $lParam)
EndFunc


Func OnKeyDown($iKeyCode)
	If $fKeyDownFlag = True Then Return
	$fKeyDownFlag = True
	Select
		Case $iKeyCode = 38
			Rotate()
		Case $iKeyCode = 40
			MoveDown()
		Case $iKeyCode = 37
			MoveLeft()
		Case $iKeyCode = 39
			MoveRight()
		Case $iKeyCode = 32
			MoveBottom()
	EndSelect
	$fKeyDownFlag = False
EndFunc


Func GenerateNewShape()
	If $iShapeIndex > 0 Then
		$iCurrentShapeIndex = $iShapeIndex
	Else
		$iCurrentShapeIndex = GenerateShapeIndex()
	EndIf
	$iCurrentShapeDirection = 0
	GenerateShapePositionArray($iCurrentShapeIndex, $iCurrentShapeDirection, $aCurrentShapePosition)
	$iShapeIndex = GenerateShapeIndex()
	If SetNewShapePosition($aCurrentShapePosition, $iCurrentShapeX, $iCurrentShapeY) = False Then
		GameOver()
	EndIf
EndFunc


Func MoveDown()
	If CheckShapePosition($aCurrentShapePosition, $iCurrentShapeX + 1, $iCurrentShapeY) = True Then
		$iCurrentShapeX = $iCurrentShapeX + 1
	Else
		AddCurrentShape($aCurrentShapePosition, $iCurrentShapeX, $iCurrentShapeY, $iCurrentShapeIndex)
		GenerateNewShape()
	EndIf
	DrawTetris()
EndFunc


Func MoveLeft()
	If CheckShapePosition($aCurrentShapePosition, $iCurrentShapeX, $iCurrentShapeY - 1) = True Then
		$iCurrentShapeY = $iCurrentShapeY - 1
		DrawTetris()
	EndIf
EndFunc


Func MoveRight()
	If CheckShapePosition($aCurrentShapePosition, $iCurrentShapeX, $iCurrentShapeY + 1) = True Then
		$iCurrentShapeY = $iCurrentShapeY + 1
		DrawTetris()
	EndIf
EndFunc


Func MoveBottom()
	If CheckShapePosition($aCurrentShapePosition, $iCurrentShapeX, $iCurrentShapeY) = True Then
		While 1
			If CheckShapePosition($aCurrentShapePosition, $iCurrentShapeX + 1, $iCurrentShapeY) = True Then
				$iCurrentShapeX = $iCurrentShapeX + 1
			Else
				ExitLoop
			EndIf
		WEnd
	EndIf
	AddCurrentShape($aCurrentShapePosition, $iCurrentShapeX, $iCurrentShapeY, $iCurrentShapeIndex)
	GenerateNewShape()
	DrawTetris()
EndFunc


Func Rotate()
	Local $iDirection, $iX, $iStartRowIndex, $fFlag, $i, $j
	$iDirection = $iCurrentShapeDirection + 1
	If $iDirection > 3 Then
		$iDirection = 0
	EndIf
	GenerateShapePositionArray($iCurrentShapeIndex, $iDirection, $aCurrentShapePosition)
	$iX = $iCurrentShapeX
	If $iCurrentShapeX < 0 Then
		$iStartRowIndex = 0
		$fFlag = False
		For $i = 0 To 3
			For $j = 0 To 3
				If $aCurrentShapePosition[$i][$j] = 1 Then
					$fFlag = True
					ExitLoop
				EndIf
			Next
			If $fFlag = True Then
				$iStartRowIndex = $i
				ExitLoop
			EndIf
		Next
		If $iX + $iStartRowIndex < 0 Then
			$iX = $iX - ($iX + $iStartRowIndex)
		EndIf
	EndIf
	If CheckShapePosition($aCurrentShapePosition,  $iX, $iCurrentShapeY) = True Then
		$iCurrentShapeDirection = $iDirection
		$iCurrentShapeX = $iX
		DrawTetris()
	Else
		GenerateShapePositionArray($iCurrentShapeIndex, $iCurrentShapeDirection, $aCurrentShapePosition)
	EndIf
EndFunc


Func GenerateShapeIndex()
	Return Random(1, 7, 1)
EndFunc


Func SetNewShapePosition($aArray, ByRef $iX, ByRef $iY)
	Local $fResult = False
	Local $iStartRowIndex, $iEndRowIndex, $i
	$iStartRowIndex = -1
	$iEndRowIndex = -1
	For $i = 0 To 3
		If $aArray[$i][0] = 1 Or $aArray[$i][1] = 1 Or $aArray[$i][2] = 1 Or $aArray[$i][3] = 1 Then
			If $iStartRowIndex < 0 Then
				$iStartRowIndex = $i
			EndIf
		Else
			If $iStartRowIndex >= 0 Then
				$iEndRowIndex = $i - 1
				ExitLoop
			EndIf
		EndIf
	Next
	$iX = 0
	$iY = 3
	For $i = $iStartRowIndex To $iEndRowIndex
		$iX = 0 - $i
		$fResult = CheckShapePosition($aArray, $iX, $iY)
		If $fResult = True Then
			ExitLoop
		EndIf
	Next
	If $fResult = False Then
		$iX = -4
	EndIf
	Return $fResult
EndFunc


Func GenerateShapePositionArray($iIndex, $iDirection, ByRef $aArray)
	Local $i, $j
	For $i = 0 To 3
		For $j = 0 To 3
			$aArray[$i][$j] = 0
		Next
	Next
	Select
		Case $iIndex = 1  ; 
			$aArray[1][1] = 1
			$aArray[1][2] = 1
			$aArray[2][1] = 1
			$aArray[2][2] = 1
		Case $iIndex = 2  ; 
			If $iDirection = 0 Or $iDirection = 2 Then
				$aArray[1][0] = 1
				$aArray[1][1] = 1
				$aArray[1][2] = 1
				$aArray[1][3] = 1
			Else
				$aArray[0][1] = 1
				$aArray[1][1] = 1
				$aArray[2][1] = 1
				$aArray[3][1] = 1
			EndIf
		Case $iIndex = 3  ; L
			Select
				Case $iDirection = 0
					$aArray[1][0] = 1
					$aArray[1][1] = 1
					$aArray[1][2] = 1
					$aArray[2][0] = 1
				Case $iDirection = 1
					$aArray[0][1] = 1
					$aArray[0][2] = 1
					$aArray[1][2] = 1
					$aArray[2][2] = 1
				Case $iDirection = 2
					$aArray[1][2] = 1
					$aArray[2][0] = 1
					$aArray[2][1] = 1
					$aArray[2][2] = 1
				Case $iDirection = 3
					$aArray[0][1] = 1
					$aArray[1][1] = 1
					$aArray[2][1] = 1
					$aArray[2][2] = 1
			EndSelect
		Case $iIndex = 4  ; L
			Select
				Case $iDirection = 0
					$aArray[1][0] = 1
					$aArray[1][1] = 1
					$aArray[1][2] = 1
					$aArray[2][2] = 1
				Case $iDirection = 1
					$aArray[0][2] = 1
					$aArray[1][2] = 1
					$aArray[2][1] = 1
					$aArray[2][2] = 1
				Case $iDirection = 2
					$aArray[1][0] = 1
					$aArray[2][0] = 1
					$aArray[2][1] = 1
					$aArray[2][2] = 1
				Case $iDirection = 3
					$aArray[0][1] = 1
					$aArray[0][2] = 1
					$aArray[1][1] = 1
					$aArray[2][1] = 1
			EndSelect
		Case $iIndex = 5  ; 
			If $iDirection = 0 Or $iDirection = 2 Then
				$aArray[1][1] = 1
				$aArray[1][2] = 1
				$aArray[2][0] = 1
				$aArray[2][1] = 1
			Else
				$aArray[0][1] = 1
				$aArray[1][1] = 1
				$aArray[1][2] = 1
				$aArray[2][2] = 1
			EndIf
		Case $iIndex = 6  ; 
			If $iDirection = 0 Or $iDirection = 2 Then
				$aArray[1][0] = 1
				$aArray[1][1] = 1
				$aArray[2][1] = 1
				$aArray[2][2] = 1
			Else
				$aArray[0][2] = 1
				$aArray[1][1] = 1
				$aArray[1][2] = 1
				$aArray[2][1] = 1
			EndIf
		Case $iIndex = 7  ; T
			Select
				Case $iDirection = 0
					$aArray[1][0] = 1
					$aArray[1][1] = 1
					$aArray[1][2] = 1
					$aArray[2][1] = 1
				Case $iDirection = 1
					$aArray[0][2] = 1
					$aArray[1][1] = 1
					$aArray[1][2] = 1
					$aArray[2][2] = 1
				Case $iDirection = 2
					$aArray[1][1] = 1
					$aArray[2][0] = 1
					$aArray[2][1] = 1
					$aArray[2][2] = 1
				Case $iDirection = 3
					$aArray[0][1] = 1
					$aArray[1][1] = 1
					$aArray[1][2] = 1
					$aArray[2][1] = 1
			EndSelect
	EndSelect
EndFunc


Func CheckShapePosition($aArray, $iX, $iY)
	Local $fResult = True
	Local $i, $j
	For $i = 0 To 3
		For $j = 0 To 3
			If $aArray[$i][$j] = 1 Then
				If $iX + $i > 17 Or $iY + $j < 0 Or $iY + $j > 9 Then
					$fResult = False
					ExitLoop
				Else
					If $iX + $i >= 0 And $aLayout[$iX + $i][$iY + $j] > 0 Then
						$fResult = False
						ExitLoop
					EndIf
				EndIf
			EndIf
		Next
	Next
	Return $fResult
EndFunc


Func AddCurrentShape($aArray, $iX, $iY, $iIndex)
	Local $i, $j, $iDeleteCount, $aRows[18], $fFlag
	For $i = 0 To 3
		For $j = 0 To 3
			If $aArray[$i][$j] = 1 And $iX + $i >= 0 And $iX + $i <= 17 _
					And $iY + $j >= 0 And $iY + $j <= 9 Then
				$aLayout[$iX + $i][$iY + $j] = $iIndex
			EndIf
		Next
	Next
	$iDeleteCount = 0
	For $i = 0 To 17
		$fFlag = True
		For $j = 0 To 9
			If $aLayout[$i][$j] = 0 Then
				$fFlag = False
				ExitLoop
			EndIf
		Next
		If $fFlag = True Then
			$iDeleteCount = $iDeleteCount + 1
			$aRows[$iDeleteCount - 1] = $i
		Else
			If $iDeleteCount > 0 And $aRows[$iDeleteCount - 1] = $i - 1 Then
				ExitLoop
			EndIf
		EndIf
	Next
	For $i = 0 To $iDeleteCount - 1
		DeleteLayoutRow($aRows[$i])
	Next
	$iRowCount = $iRowCount + $iDeleteCount
	Select
		Case $iDeleteCount = 1
			$iScore = $iScore + 1
		Case $iDeleteCount = 2
			$iScore = $iScore + 3
		Case $iDeleteCount = 3
			$iScore = $iScore + 5
		Case $iDeleteCount = 4
			$iScore = $iScore + 7
	EndSelect
EndFunc


Func DeleteLayoutRow($iRowIndex)
	Local $i, $j
	For $i = $iRowIndex - 1 To 0 Step -1
		For $j = 0 To 9
			$aLayout[$i + 1][$j] = $aLayout[$i][$j]
		Next
	Next
	If $iRowIndex > 0 Then
		For $i = 0 To 9
			$aLayout[0][$i] = 0
		Next
	EndIf
EndFunc


Func SaveResult($sName)
	Local $oConnection, $sSQL
	CreateDatabase()
	$oConnection = ObjCreate("ADODB.Connection")
	$oConnection.Open($ACCESS_CONNECTION)
	$sSQL = "INSERT INTO LOGS (USER_NAME, LAST_DATE, SECONDS, ROW_COUNT, SCORE) VALUES (" & _
			"'" & StringReplace(StringStripWS($sName, 3), "'", "''") & "', " & _
			"'" & _NowCalc() & "', " & _
			$iTotalSeconds & ", " & _
			$iRowCount & ", " & _
			$iScore & ")"
	$oConnection.Execute($sSQL)
	$oConnection.Close()
	$oConnection = False
EndFunc


Func CreateDatabase()
	Local $ADOXCatalog, $aSQL[4], $i
	If FileExists($DATABASE_FILENAME) = 1 Then Return
	$ADOXCatalog = ObjCreate("ADOX.Catalog")
	$aSQL[0] = "CREATE TABLE [LOGS] (" & _
			"[USER_NAME] NVARCHAR(50) NOT NULL, " & _
			"[LAST_DATE] NVARCHAR(20) NOT NULL, " & _
			"[SECONDS] INTEGER NOT NULL,  " & _
			"[ROW_COUNT] INTEGER NOT NULL,  " & _
			"[SCORE] INTEGER NOT NULL)"
	$aSQL[1] = "CREATE INDEX [SCORE_IDX] ON [LOGS]([SCORE] DESC)"
	$aSQL[2] = "CREATE INDEX [USER_NAME_IDX] ON [LOGS]([USER_NAME] ASC)"
	$aSQL[3] = "CREATE INDEX [LAST_DATE_IDX] ON [LOGS]([LAST_DATE] ASC)"
	$ADOXCatalog.Create($ACCESS_CONNECTION)
	For $i = 0 To 3
		$ADOXCatalog.ActiveConnection.Execute($aSQL[$i])
	Next
	$ADOXCatalog = False
EndFunc


Func SetGameData($hListView)
	Local $oConnection, $sSQL, $oRecordSet, $iIndex
	CreateDatabase()
	$oConnection = ObjCreate("ADODB.Connection")
	$oConnection.Open($ACCESS_CONNECTION)
	$sSQL = "SELECT TOP 20 USER_NAME, SCORE, ROW_COUNT, SECONDS, LAST_DATE " & _
			"FROM LOGS ORDER BY SCORE DESC, USER_NAME ASC, LAST_DATE ASC"
	$oRecordSet = $oConnection.Execute($sSQL)
	$iIndex = 0
	While $oRecordSet.BOF = False And $oRecordSet.EOF = False
		$iIndex = $iIndex + 1
		GUICtrlCreateListViewItem($iIndex & "|" & StringStripWS($oRecordSet.Fields.Item(0).Value, 3) & "|" & _
				$oRecordSet.Fields.Item(1).Value & "|" & $oRecordSet.Fields.Item(2).Value & "|" & _
				CalcTimeCost($oRecordSet.Fields.Item(3).Value) & "|" & _
				StringStripWS($oRecordSet.Fields.Item(4).Value, 3), _
				$hListView)
		$oRecordSet.MoveNext()
	WEnd
	$oRecordSet.Close()
	$oRecordSet = False
	$oConnection.Close()
	$oConnection = False
EndFunc


Func SetGameData2($hListView)
	Local $oConnection, $sSQL, $oRecordSet, $iIndex
	CreateDatabase()
	$oConnection = ObjCreate("ADODB.Connection")
	$oConnection.Open($ACCESS_CONNECTION)
	$sSQL = "SELECT TOP 20 * FROM " & _
			"[SELECT USER_NAME, AVG(SCORE) AS AVG_SCORE, SUM(SCORE) AS SUM_SCORE, " & _
			"SUM(ROW_COUNT) AS SUM_ROW_COUNT, SUM(SECONDS) AS SUM_SECONDS " & _
			"FROM LOGS GROUP BY USER_NAME]. AS A " & _
			"ORDER BY AVG_SCORE DESC, USER_NAME ASC"
	$oRecordSet = $oConnection.Execute($sSQL)
	$iIndex = 0
	While $oRecordSet.BOF = False And $oRecordSet.EOF = False
		$iIndex = $iIndex + 1
		GUICtrlCreateListViewItem($iIndex & "|" & StringStripWS($oRecordSet.Fields.Item(0).Value, 3) & "|" & _
				$oRecordSet.Fields.Item(1).Value & "|" & $oRecordSet.Fields.Item(2).Value & "|" & _
				$oRecordSet.Fields.Item(3).Value & "|" & CalcTimeCost($oRecordSet.Fields.Item(4).Value), _
				$hListView)
		$oRecordSet.MoveNext()
	WEnd
	$oRecordSet.Close()
	$oRecordSet = False
	$oConnection.Close()
	$oConnection = False
EndFunc


Func _WinAPI_LockWindowUpdate($hWnd)
	Local $aResult = DllCall("user32.dll", "int", "LockWindowUpdate", "handle", $hWnd)
	If @error Then Return SetError(@error, @extended, 0)
	Return $aResult[0]
EndFunc